/* dragon-watch-applet: dragon-watch-applet.cc
 *
 * Copyright 2005 Chun-Chung Chen <cjj@u.washington.edu>
 *
 * This program is free software; you can redistribute it and'or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <libsoup/soup.h>
#include <gtkmm/box.h>
#include <gtkmm/main.h>
#include <gtkmm/stock.h>
#include <gtkmm/entry.h>
#include <gtkmm/dialog.h>
#include <gtkmm/window.h>
#include <gtkmm/tooltips.h>
#include <gtkmm/spinbutton.h>
#include <gtkmm/drawingarea.h>
#include <libpanelappletmm/applet.h>
#include <libpanelappletmm/init.h>
#include <libpanelappletmm/factory.h>
#include <libglademm/xml.h>
#include <iostream>
#include <sstream>
#include <fstream>
#include <cstdio>

#include "dragon-icon.inl"
#include "dragon-pref.inl"
#include "applet-xml.inl"

class DragonSoup
{
public:
	enum State {
		STATE_INIT,
		STATE_LOGIN,
		STATE_LOGIN_OK,
		STATE_LOGIN_ERR,
		STATE_QUERY,
		STATE_QUERY_OK,
		STATE_QUERY_ERR
	};

	DragonSoup();
	~DragonSoup();
	void setup(std::string uid, std::string pass);
	void login();
	void query();
	State get_state();
	sigc::signal<void> replied;

	bool get_status();
	int get_ngames();
	int get_nmessages();
private:
	// authentication information
	std::string userid;
	std::string passwd;

	// connection
	SoupSession * soup;
	std::string handle;
	std::string sessioncode;

	State state;

	static void login_receive(SoupMessage * msg, gpointer data);
	static void query_receive(SoupMessage * msg, gpointer data);

	std::vector<std::string> status_text;
	int num_games;
	int num_messages;
	bool parse_status();
};

DragonSoup::DragonSoup() :
	state(STATE_INIT)
{
	if (! g_thread_supported()) g_thread_init(0); // for libsoup
	soup = soup_session_async_new();
}

DragonSoup::~DragonSoup()
{
	g_object_unref(soup);
}

void DragonSoup::setup(std::string uid, std::string pass)
{
	userid = uid;
	passwd = pass;
	if (state == STATE_LOGIN && state == STATE_QUERY) soup_session_abort(soup);
	state = STATE_INIT;
	handle.clear();
	sessioncode.clear();
}

void DragonSoup::login()
{
	if (state == STATE_LOGIN && state == STATE_QUERY) soup_session_abort(soup);
	std::ostringstream os;
	os << "http://www.dragongoserver.net/login.php?userid=" << userid
	   << "&passwd=" << passwd;
	SoupMessage * msg = soup_message_new(SOUP_METHOD_POST, os.str().c_str());
	soup_message_set_flags(msg, SOUP_MESSAGE_NO_REDIRECT);
	soup_session_queue_message(soup, msg, login_receive, this);
	state = STATE_LOGIN;
}

void DragonSoup::query()
{
	if (state == STATE_LOGIN && state == STATE_QUERY) soup_session_abort(soup);
	std::ostringstream os;
	os << "handle=" << handle << "; sessioncode=" << sessioncode;
	SoupMessage * msg = soup_message_new(SOUP_METHOD_GET, "http://www.dragongoserver.net/quick_status.php");
	soup_message_add_header(msg->request_headers, "Cookie", os.str().c_str());
	soup_message_set_flags(msg, SOUP_MESSAGE_NO_REDIRECT);
	soup_session_queue_message(soup, msg, query_receive, this);
	state = STATE_QUERY;
}

bool DragonSoup::get_status()
{
	return num_games || num_messages;
}

int DragonSoup::get_ngames()
{
	return num_games;
}

int DragonSoup::get_nmessages()
{
	return num_messages;
}

DragonSoup::State DragonSoup::get_state()
{
	return state;
}

void DragonSoup::login_receive(SoupMessage * msg, gpointer data)
{
	DragonSoup * me = static_cast<DragonSoup *>(data);
	const GSList * gsl = soup_message_get_header_list(msg->response_headers, "Set-Cookie");
	me->handle.clear();
	me->sessioncode.clear();
	for (const GSList * i = gsl; i; i = i->next) {
		std::string line = (const char *) i->data;
		if (line.substr(0, 7) == "handle=") {
			int d = line.find(';', 7);
			me->handle = line.substr(7, d - 7);
		}
		else if (line.substr(0, 12) == "sessioncode=") {
			int d = line.find(';', 12);
			me->sessioncode = line.substr(12, d - 12);
		}
	}
	if (me->handle.size() && me->sessioncode.size()) me->state = STATE_LOGIN_OK;
	else me->state = STATE_LOGIN_ERR;
	me->replied();
}

void DragonSoup::query_receive(SoupMessage * msg, gpointer data)
{
	DragonSoup * me = static_cast<DragonSoup *>(data);
	if (SOUP_STATUS_IS_SUCCESSFUL(msg->status_code)) {
		me->status_text.clear();
		std::string line;
		char * b = msg->response.body;
		for (unsigned i = 0; i < msg->response.length; i ++) {
			if (b[i] == '\r' || b[i] == '\n') {
				if (line != "") me->status_text.push_back(line);
				line = "";
			}
			else line += b[i];
		}
		if (line != "") me->status_text.push_back(line);
		if (me->status_text.size() && me->parse_status()) me->state = STATE_QUERY_OK;
		else me->state = STATE_QUERY_ERR;
	}
	else me->state = STATE_QUERY_ERR;
	me->replied();
}

bool DragonSoup::parse_status()
{
	num_games = 0;
	num_messages = 0;
	for (std::vector<std::string>::iterator i = status_text.begin(); i != status_text.end(); i ++) {
		unsigned pos = 0;
		pos = i->find('\'', 0);
		if (pos == std::string::npos) {
			if (* i != "Warning: nothing found") return false;
			continue;
		}
		pos ++;
		if (pos >= i->length()) return false;
		if ((* i)[pos] == 'G') num_games ++;
		else if ((* i)[pos] == 'M') num_messages ++;
		else return false;
	}
	return true;
}

class DragonIcon :
	public Gtk::DrawingArea
{
public:
	enum State {
		STATE_NORMAL,
		STATE_ACTIVE,
		STATE_ERROR
	};
	DragonIcon();

	void set_state(State s);
	double aspect_ratio();
	const Glib::RefPtr<Gdk::Pixbuf> & get_icon();
private:
	State state;

	Glib::RefPtr<Gdk::Pixbuf> icon;
	Glib::RefPtr<Gdk::Pixbuf> scaled_icon;

	Glib::RefPtr<Gdk::Window> win;
	Glib::RefPtr<Gtk::Style> style;
	int win_w;
	int win_h;
	Glib::RefPtr<Gdk::GC> gc;
	void draw_icon();
	void on_realize();
	bool on_configure_event(GdkEventConfigure * event);
	void on_style_changed(const Glib::RefPtr<Gtk::Style> & prev_style);
	bool on_expose_event(GdkEventExpose * event);
};

DragonIcon::DragonIcon() :
	state(STATE_ERROR)
{
	icon = Gdk::Pixbuf::create_from_inline(- 1, dragon_icon);
}

void DragonIcon::set_state(State s)
{
	if (s != state) {
		state = s;
		draw_icon();
	}
}

double DragonIcon::aspect_ratio()
{
	return icon->get_height() / double(icon->get_width());
}

const Glib::RefPtr<Gdk::Pixbuf> & DragonIcon::get_icon()
{
	return icon;
}

void DragonIcon::draw_icon()
{
	if (! is_realized()) return;
	win->clear_area(0, 0, win_w, win_h);
	if (! gc) gc = Gdk::GC::create(win);
	switch (state) {
	case STATE_NORMAL:
		gc->set_rgb_fg_color(Gdk::Color("white"));
		break;
	case STATE_ACTIVE:
		gc->set_rgb_fg_color(Gdk::Color("black"));
		break;
	case STATE_ERROR:
		gc->set_rgb_fg_color(Gdk::Color("red"));
		break;
	}
	win->draw_arc(gc, true, win_w / 12, win_h / 12, win_w * 5 / 6, win_h * 5 / 6, 0, 64 * 360);
	win->draw_pixbuf(style->get_fg_gc(get_state()), scaled_icon, 0, 0, 0, 0, win_w, win_h, Gdk::RGB_DITHER_NORMAL, 0, 0);
}

void DragonIcon::on_realize()
{
	DrawingArea::on_realize();
	win = get_window();
	style = get_style();
}

bool DragonIcon::on_configure_event(GdkEventConfigure * event)
{
	win_w = event->width;
	win_h = event->height;
	scaled_icon = icon->scale_simple(win_w, win_h, Gdk::INTERP_HYPER);
	return true;
}

void DragonIcon::on_style_changed(const Glib::RefPtr<Gtk::Style> & prev_style)
{
	DrawingArea::on_style_changed(prev_style);
	style = get_style();
}

bool DragonIcon::on_expose_event(GdkEventExpose * event)
{
	if (! event->count) draw_icon();
	return true;
}

class DragonApplet :
	public Gnome::Panel::Applet
{
public:
	explicit DragonApplet(PanelApplet * castitem);
	virtual ~DragonApplet();
private:
	DragonIcon di;
	void on_size_allocate(Gtk::Allocation & allocation);
	bool on_button_press_event(GdkEventButton * event);

	DragonSoup ds;
	Gtk::Dialog * pref_dialog;
	Gtk::Entry * userid_entry;
	Gtk::Entry * passwd_entry;
	Gtk::SpinButton * interval_entry;
	Gtk::Entry * command_entry;
	Gtk::Tooltips tooltip;

	static void menu_update(BonoboUIComponent *, void * applet, const gchar *);
	static void menu_preferences(BonoboUIComponent *, void * applet, const gchar *);
	void start_watch();
	void set_tooltip(const std::string & message);

	sigc::connection conn_periodic;
	bool proc_periodic();
	void dragon_reply();
};

DragonApplet::DragonApplet(PanelApplet * castitem) :
	Gnome::Panel::Applet(castitem)
{
	static const BonoboUIVerb menu_verbs[] = {
		BONOBO_UI_VERB("Update", & DragonApplet::menu_update),
		BONOBO_UI_VERB("Preferences", & DragonApplet::menu_preferences),
		BONOBO_UI_VERB_END
	};
	setup_menu(applet_xml, menu_verbs, this);
	add_preferences("/schemas/apps/dragon_watch_applet/prefs");
	set_flags(Gnome::Panel::APPLET_EXPAND_MINOR);

	set_border_width(0);
	add(di);
	// try to find the best initial size
	int sz = get_size();
	int rw = sz;
	int rh = sz;
	if (di.aspect_ratio() < 1) rh = int((sz - 2) * di.aspect_ratio() + 2.5);
	else rw = int((sz - 2) / di.aspect_ratio() + 2.5);
	set_size_request(rw, rh);

	show_all();
	ds.replied.connect(sigc::mem_fun(* this, & DragonApplet::dragon_reply));
	start_watch();
}

DragonApplet::~DragonApplet()
{
}

void DragonApplet::on_size_allocate(Gtk::Allocation & allocation)
{
	Gnome::Panel::Applet::on_size_allocate(allocation);
	// try to maximize the size
	int a_w = allocation.get_width();
	int a_h = allocation.get_height();
	int p_w = int((a_h - 2) / di.aspect_ratio() + 2.5);
	int p_h = int((a_w - 2) * di.aspect_ratio() + 2.5);
	int rw = a_w + 1;
	int rh = a_h + 1;
	if (p_w > a_w) rh = p_h;
	else if (p_h > a_h) rw = p_w;
	else return;
	set_size_request(rw, rh);
}

bool DragonApplet::on_button_press_event(GdkEventButton * event)
{
	if (event->button != 1) return Gnome::Panel::Applet::on_button_press_event(event);
	std::string cmd = gconf_get_string("command");
	if (cmd.length()) Glib::spawn_command_line_async(cmd);
	return true;
}

void DragonApplet::menu_update(BonoboUIComponent *, void * applet, const gchar *)
{
	DragonApplet * me = static_cast<DragonApplet *>(applet);
	switch (me->ds.get_state()) {
	case DragonSoup::STATE_LOGIN_OK:
	case DragonSoup::STATE_QUERY_OK:
		if (me->conn_periodic) me->conn_periodic.disconnect();
		me->ds.query();
		me->conn_periodic = Glib::signal_timeout().connect(sigc::mem_fun(* me, & DragonApplet::proc_periodic), me->gconf_get_int("interval") * 1000);
		break;
	default:
		me->start_watch();
		break;
	}
}

void DragonApplet::menu_preferences(BonoboUIComponent *, void * applet, const gchar *)
{
	DragonApplet * me = static_cast<DragonApplet *>(applet);
	if (! me->pref_dialog) {
		Glib::RefPtr<Gnome::Glade::Xml> xml = Gnome::Glade::Xml::create_from_buffer(dragon_pref.c_str(), dragon_pref.length());
		me->pref_dialog = dynamic_cast<Gtk::Dialog *>(xml->get_widget("pref_dialog"));
		me->userid_entry = dynamic_cast<Gtk::Entry *>(xml->get_widget("userid_entry"));
		me->passwd_entry = dynamic_cast<Gtk::Entry *>(xml->get_widget("passwd_entry"));
		me->interval_entry = dynamic_cast<Gtk::SpinButton *>(xml->get_widget("interval_entry"));
		me->command_entry = dynamic_cast<Gtk::Entry *>(xml->get_widget("command_entry"));
		me->pref_dialog->set_icon(me->di.get_icon());
	}
	// setup entry fields
	me->userid_entry->set_text(me->gconf_get_string("userid"));
	me->passwd_entry->set_text(me->gconf_get_string("passwd"));
	me->interval_entry->set_value(me->gconf_get_int("interval"));
	me->command_entry->set_text(me->gconf_get_string("command"));
	me->pref_dialog->present();
	if (me->pref_dialog->run() == Gtk::RESPONSE_OK) {
		me->gconf_set("userid", me->userid_entry->get_text());
		me->gconf_set("passwd", me->passwd_entry->get_text());
		me->gconf_set("interval", int(me->interval_entry->get_value()));
		me->gconf_set("command", me->command_entry->get_text());
		me->start_watch();
	}
	me->pref_dialog->hide();
}

void DragonApplet::start_watch()
{
	std::string userid = gconf_get_string("userid");
	std::string passwd = gconf_get_string("passwd");
	ds.setup(userid, passwd);
	if (conn_periodic) conn_periodic.disconnect();
	if (userid.length() && passwd.length()) {
		ds.login();
		conn_periodic = Glib::signal_timeout().connect(sigc::mem_fun(* this, & DragonApplet::proc_periodic), gconf_get_int("interval") * 1000);
		set_tooltip(_("Authenticating..."));
	} else set_tooltip(_("Need to setup login preferences."));
	di.set_state(DragonIcon::STATE_ERROR);
}

void DragonApplet::set_tooltip(const std::string & message)
{
	tooltip.set_tip(* this, message);
}

bool DragonApplet::proc_periodic()
{
	switch (ds.get_state()) {
	case DragonSoup::STATE_LOGIN_OK:
	case DragonSoup::STATE_QUERY_OK:
	case DragonSoup::STATE_QUERY_ERR:
		ds.query();
		set_tooltip(_("Querying..."));
		break;
	case DragonSoup::STATE_LOGIN_ERR:
		ds.login();
		set_tooltip(_("Authenticating..."));
		break;
	case DragonSoup::STATE_LOGIN:
	case DragonSoup::STATE_QUERY:
		// still waiting...
		break;
	default:
		set_tooltip(_("Internal error!")); // impossible situation!
	}
	return true;
}

void DragonApplet::dragon_reply()
{
	switch (ds.get_state()) {
	case DragonSoup::STATE_LOGIN_OK:
		// immediate query follows successful login
		if (conn_periodic) conn_periodic.disconnect();
		ds.query();
		conn_periodic = Glib::signal_timeout().connect(sigc::mem_fun(* this, & DragonApplet::proc_periodic), gconf_get_int("interval") * 1000);
		di.set_state(DragonIcon::STATE_ERROR);
		set_tooltip(_("Querying..."));
		break;
	case DragonSoup::STATE_QUERY_OK:
		if (ds.get_status()) {
			di.set_state(DragonIcon::STATE_ACTIVE);
			char buf[100];
			snprintf(buf, 100, _("Have %d game(s), %d message(s)."), ds.get_ngames(), ds.get_nmessages());
			set_tooltip(buf);
		}
		else {
			di.set_state(DragonIcon::STATE_NORMAL);
			set_tooltip(_("Have no games nor messages."));
		}
		break;
	case DragonSoup::STATE_LOGIN_ERR:
		di.set_state(DragonIcon::STATE_ERROR);
		set_tooltip(_("Fail to login!"));
		break;
	case DragonSoup::STATE_QUERY_ERR:
		di.set_state(DragonIcon::STATE_ERROR);
		set_tooltip(_("Fail to get status!"));
		break;
	default: // impossible situation
		di.set_state(DragonIcon::STATE_ERROR);
		set_tooltip(_("Internal error!"));
	}
}

int main(int argc, char * argv[])
{
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	Gnome::Panel::init("DragonWatchApplet", VERSION, argc, argv);

	try {
		int r = Gnome::Panel::factory_main<DragonApplet>("OAFIID:DragonWatch_Factory");
		return r;
	}
	catch (const Glib::Error & err) {
		std::cerr << err.what() << std::endl;
		return 0;
	}
}
